//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "uploadqueue.h"
#include "packets.h"
#include "emule.h"
#include "knownfile.h"
#include "listensocket.h"
#include "ini2.h"
#include "math.h"
#include "Exceptions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


static uint32 counter, sec,statsave;
// -khaos--+++> Added iupdateconnstats...
static uint32 igraph, istats, iupdateconnstats;
// <-----khaos-

//TODO rewrite the whole networkcode, use overlapped sockets

CUploadQueue::CUploadQueue(CPreferences* in_prefs){
	app_prefs = in_prefs;
	h_timer = SetTimer(0,141,100,UploadTimer);
	if (!h_timer)
		theApp.emuledlg->AddDebugLogLine(false,GetResString(IDS_ERR_TIMERCREATEFAILED));
	estadatarate = 2000;
	datarate = 0;
	dataratems = 0;
	datarateave = 0;
	bannedcount = 0;
	counter=0;
	successfullupcount = 0;
	failedupcount = 0;
	totaluploadtime = 0;
	m_nUpDataRateMSOverhead = 0;
	m_nUpDatarateOverhead = 0;
	m_nUpDataOverheadSourceExchange = 0;
	m_nUpDataOverheadFileRequest = 0;
	m_nUpDataOverheadOther = 0;
	m_nUpDataOverheadServer = 0;
	m_nUpDataOverheadSourceExchangePackets = 0;
	m_nUpDataOverheadFileRequestPackets = 0;
	m_nUpDataOverheadOtherPackets = 0;
	m_nUpDataOverheadServerPackets = 0;
	m_nLastStartUpload = 0;
	statsave=0;
	// -khaos--+++>
	iupdateconnstats=0;
	// <-----khaos-
	// By BadWolf - Accurate Speed Measurement
	sumavgUDRO = 0;
	sendperclient = 0;
	uLastAcceptNewClient = ::GetTickCount();
	sumavgdata = 0;
	m_delay = ::GetTickCount();
	m_delaytmp = 0;
	// END By BadWolf - Accurate Speed Measurement
}

void CUploadQueue::AddUpNextClient(CUpDownClient* directadd){
	POSITION toadd = 0;
	POSITION toaddlow = 0;
	uint32	bestscore = 0;
	uint32  bestlowscore = 0;
	CUpDownClient* newclient;
	// select next client or use given client
	if (!directadd){
		POSITION pos1, pos2;
		for (pos1 = waitinglist.GetHeadPosition();( pos2 = pos1 ) != NULL;){
			waitinglist.GetNext(pos1);
			CUpDownClient* cur_client =	waitinglist.GetAt(pos2);
			// clear dead clients
			ASSERT ( cur_client->GetLastUpRequest() );
			if ((::GetTickCount() - cur_client->GetLastUpRequest() > MAX_PURGEQUEUETIME) || !theApp.sharedfiles->GetFileByID(cur_client->GetUploadFileID()) ){
				RemoveFromWaitingQueue(pos2,true);	
				if (!cur_client->socket)
					cur_client->Disconnected();
				continue;
			}
			// finished clearing
			uint32 cur_score = cur_client->GetScore(true);
			if ( cur_score > bestscore){
				bestscore = cur_score;
				toadd = pos2;
			} else {
				cur_score = cur_client->GetScore(false);
				if ((cur_score > bestlowscore) && !cur_client->AddNextConnect){
					bestlowscore = cur_score;
					toaddlow = pos2;
				}
			}
		}
		
		if (bestlowscore > bestscore){
			newclient = waitinglist.GetAt(toaddlow);
			newclient->AddNextConnect = true;
		}
		if (!toadd)
			return;
		newclient = waitinglist.GetAt(toadd);
		lastupslotHighID = true; // VQB LowID alternate
		//theApp.emuledlg->AddLogLine(true,"Added High ID: " + (CString)newclient->GetUserName()); // VQB:  perhaps only add to debug log?
		RemoveFromWaitingQueue(toadd, true);
		theApp.emuledlg->transferwnd.ShowQueueCount(waitinglist.GetCount());
	}
	else {
		newclient = directadd;
		/*if (!IsDownloading(newclient)){
            if (newclient->HasLowID())
                theApp.emuledlg->AddLogLine(true,"DirectAdd:  LowID: " + (CString)newclient->GetUserName());
			else
                theApp.emuledlg->AddLogLine(true,"DirectAdd:  HighID: " + (CString)newclient->GetUserName());
		}*/
	}

	if (IsDownloading(newclient)){
		return;
	}
	// tell the client that we are now ready to upload
	if (!newclient->socket || !newclient->socket->IsConnected()){
		newclient->SetUploadState(US_CONNECTING);
		newclient->TryToConnect(true);
	}
	else{
		Packet* packet = new Packet(OP_ACCEPTUPLOADREQ,0);
		theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
		newclient->socket->SendPacket(packet,true);
		newclient->SetUploadState(US_UPLOADING);
	}
	newclient->SetUpStartTime();
	newclient->ResetSessionUp();
	uploadinglist.AddTail(newclient);
	
	// statistic
	CKnownFile* reqfile = theApp.sharedfiles->GetFileByID((uchar*)newclient->GetUploadFileID());
	if (reqfile){
		reqfile->statistic.AddAccepted();
	}
		
	theApp.emuledlg->transferwnd.uploadlistctrl.AddClient(newclient);

}

void CUploadQueue::Process() {
	while ( avarage_dr_list.GetCount()>150)
		dataratems-=avarage_dr_list.RemoveHead().datalen;

	uint32 tick_avg = 100;
	if (avarage_dr_list.GetCount() >2) {
		uint32 deltat=(avarage_dr_list.GetTail().timestamp - avarage_dr_list.GetHead().timestamp);
		if (deltat>0) datarate = (1024*dataratems) / deltat ;
		tick_avg = deltat / avarage_dr_list.GetCount();
	} else
		datarate = 0;

	if (AcceptNewClient() && waitinglist.GetCount()){
		m_nLastStartUpload = ::GetTickCount();
		AddUpNextClient();
	}

	if (!uploadinglist.GetCount()) {
		return;
	}

	int16 clientsrdy = 0;
	for (POSITION pos = uploadinglist.GetHeadPosition();pos != 0;uploadinglist.GetNext(pos)){
		CUpDownClient* cur_client = uploadinglist.GetAt(pos);
		if ( (cur_client->socket) && (!cur_client->socket->IsBusy()) && cur_client->HasBlocks())
			clientsrdy++;
	}

	uint32 sendperclient;

	// [Elandal]
	if (!clientsrdy) {
		if (estadatarate >200) estadatarate -= 200; else estadatarate =100;
		sendperclient = estadatarate;
		//estadatarate_tmp= estadatarate;
	} else{
		uint32 estadatarate_tmp;
		uint32 maxupload_real= (app_prefs->GetMaxUpload()==UNLIMITED)?app_prefs->GetMaxGraphUploadRate():app_prefs->GetMaxUpload();
		uint32 maxupload = maxupload_real * 2; // 2% of MaxUpload in bytes / tick
		if(maxupload<200) maxupload = 200;
		estadatarate += maxupload;

		// Patch by BadWolf - Accurate Upload Datarate
		DWORD deltmp = ::GetTickCount();
		if (m_delaytmp==0) m_delaytmp=deltmp-100;
		m_delay = deltmp - m_delaytmp ; // we've lost 3 millisecs somewhere
		m_delaytmp = deltmp;

		estadatarate_tmp = estadatarate / 100 * m_delay;
		uint32 maxrate = maxupload_real * m_delay;
		uint32 maxrate_normalized = maxupload_real * 100;
		if (estadatarate > maxrate_normalized) {
			estadatarate_tmp = maxrate;
			estadatarate = maxrate_normalized;
		}
		sendperclient = estadatarate_tmp/clientsrdy;
	}

	uint32 datarateX =0;
	POSITION pos1,pos2;
	for (pos1 = uploadinglist.GetHeadPosition();( pos2 = pos1 ) != NULL; ){
		uploadinglist.GetNext(pos1);
		CUpDownClient* cur_client = uploadinglist.GetAt(pos2);
		datarateX += cur_client->SendBlockData(sendperclient);
	}
	TransferredData newitem = {datarateX, ::GetTickCount()};  // theApp.stat_sessionSentBytes
	avarage_dr_list.AddTail(newitem);
	dataratems+=datarateX;
};

bool CUploadQueue::AcceptNewClient(){
	// check if we can allow a new client to start downloading form us
	if (::GetTickCount() - m_nLastStartUpload < 1000 && datarate < 102400 )
		return false;
	if (uploadinglist.GetCount() < MIN_UP_CLIENTS_ALLOWED)
		return true;
	else if (uploadinglist.GetCount() >= MAX_UP_CLIENTS_ALLOWED)
		return false;

	uint32	upPerClient = UPLOAD_CLIENT_DATARATE + datarate/50;
	if( upPerClient > 11000 )
		upPerClient = 11000;

	//now the final check
	if (theApp.glob_prefs->GetMaxUpload() == UNLIMITED){
		if ((uint32)uploadinglist.GetCount() < ((datarate/upPerClient)+2))
			return true;
	}
	else{
		uint16 nMaxSlots = 0;
		if (theApp.glob_prefs->GetMaxUpload() > 10){
			nMaxSlots += 2;
			nMaxSlots += (uint16)ceil((float)((theApp.glob_prefs->GetMaxUpload() - 10)*1024) / upPerClient);
		}
		else
			nMaxSlots = MIN_UP_CLIENTS_ALLOWED;
//theApp.emuledlg->AddLogLine(true,"maxslots=%u, upPerClient=%u",nMaxSlots,upPerClient);
		if ((uint32)uploadinglist.GetCount() < (datarate/UPLOAD_CHECK_CLIENT_DR) && uploadinglist.GetCount() <= nMaxSlots )
			return true;
	}
	//nope
	return false;
}

CUploadQueue::~CUploadQueue(){
	KillTimer(0,141);
}

CUpDownClient* CUploadQueue::GetWaitingClientByIP_UDP(uint32 dwIP, uint16 nUDPPort){
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;){
		CUpDownClient* cur_client = waitinglist.GetNext(pos);
		if (dwIP == cur_client->GetIP() && nUDPPort == cur_client->GetUDPPort())
			return cur_client;
	}
	return 0;
}

#ifdef _DEBUG
CUpDownClient* CUploadQueue::GetWaitingClientByIP(uint32 dwIP){
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;){
		CUpDownClient* cur_client = waitinglist.GetNext(pos);
		if (dwIP == cur_client->GetIP())
			return cur_client;
	}
	return 0;
}
#endif

void CUploadQueue::UpdateBanCount(){
	int count=0;
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;waitinglist.GetNext(pos)){
		CUpDownClient* cur_client= waitinglist.GetAt(pos);
		if(cur_client->IsBanned())
			count++;
	}
	SetBanCount(count);
}
void CUploadQueue::AddClientToQueue(CUpDownClient* client, bool bIgnoreTimelimit){
	if (theApp.serverconnect->IsConnected() && theApp.serverconnect->IsLowID()
		&& !theApp.serverconnect->IsLocalServer(client->GetServerIP(),client->GetServerPort())
		&& client->GetDownloadState() == DS_NONE && !client->IsFriend()
		&& GetWaitingUserCount() > 50)
		return;
	client->AddAskedCount();
	client->SetLastUpRequest();
	if (!bIgnoreTimelimit){
		if (client->IsBanned()){
			if (::GetTickCount() - client->GetBanTime() > 18000000){
				client->UnBan();
			}
			else
				return;
		}
		client->AddRequestCount(client->GetUploadFileID());
	}
	// check for double
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;waitinglist.GetNext(pos)){
		CUpDownClient* cur_client= waitinglist.GetAt(pos);
		if (cur_client == client){	//already on queue
// VQB LowID Slot Patch -- note:  should add limit so only if #slots < UL -or- UL+1 for Low UL (?)
			if (client->AddNextConnect && (uploadinglist.GetCount() < theApp.glob_prefs->GetMaxUpload())){
				if (lastupslotHighID) {
					client->AddNextConnect = false;
					RemoveFromWaitingQueue(client, true);
					AddUpNextClient(client);
					lastupslotHighID = false; // LowID alternate
					// theApp.emuledlg->AddLogLine(true,"Added Low ID User On Reconnect: " + (CString)client->GetUserName()); // VQB:  perhaps only add to debug log?
				return;
			}
				else
					;// theApp.emuledlg->AddLogLine(true, "Skipped LowID User on Reconnect: " + (CString)client->GetUserName());
			}
// VQB end
			client->SendRankingInfo();
			theApp.emuledlg->transferwnd.queuelistctrl.RefreshClient(client);
			return;			
		}
		else if ( client->Compare(cur_client) ) {
			// another client with same ip or hash
			theApp.emuledlg->AddDebugLogLine(false,CString(GetResString(IDS_SAMEUSERHASH)),client->GetUserName(),cur_client->GetUserName(),cur_client->GetUserName() );
			RemoveFromWaitingQueue(pos,true);	
			if (!cur_client->socket)
				cur_client->Disconnected();	
			return;
		}
	}
	// done

	// Add clients server to list.
	if (theApp.glob_prefs->AddServersFromClient()){
		in_addr host;
		host.S_un.S_addr = client->GetServerIP();
		CServer* srv = new CServer(client->GetServerPort(), inet_ntoa(host));
		srv->SetListName(srv->GetAddress());

		if (!theApp.emuledlg->serverwnd.serverlistctrl.AddServer(srv, true))
			delete srv;
		/*else
			theApp.emuledlg->AddLogLine(false,"Added new server: %s:%d", srv->GetFullIP(), srv->GetPort());*/
	}

	// statistic values
	CKnownFile* reqfile = theApp.sharedfiles->GetFileByID((uchar*)client->GetUploadFileID());
		if (reqfile)
			reqfile->statistic.AddRequest();
	// TODO find better ways to cap the list
		if ((uint32)waitinglist.GetCount() > (theApp.glob_prefs->GetQueueSize()+bannedcount))
		return;
	if (client->IsDownloading()){
		// he's already downloading and wants probably only another file
		Packet* packet = new Packet(OP_ACCEPTUPLOADREQ,0);
		theApp.uploadqueue->AddUpDataOverheadFileRequest(packet->size);
		client->socket->SendPacket(packet,true);
		return;
	}
	if (waitinglist.IsEmpty() && AcceptNewClient()){
		AddUpNextClient(client);
		m_nLastStartUpload = ::GetTickCount();
	}
	else{
		waitinglist.AddTail(client);
		client->SetUploadState(US_ONUPLOADQUEUE);
		client->SendRankingInfo();
		theApp.emuledlg->transferwnd.queuelistctrl.AddClient(client);
		theApp.emuledlg->transferwnd.ShowQueueCount(waitinglist.GetCount());
	}
}

bool CUploadQueue::RemoveFromUploadQueue(CUpDownClient* client, bool updatewindow){
	for (POSITION pos = uploadinglist.GetHeadPosition();pos != 0;uploadinglist.GetNext(pos)){
		if (client == uploadinglist.GetAt(pos)){
			if (updatewindow)
				theApp.emuledlg->transferwnd.uploadlistctrl.RemoveClient(uploadinglist.GetAt(pos));
			uploadinglist.RemoveAt(pos);
			if( client->GetTransferedUp() ){
				successfullupcount++;
				totaluploadtime += client->GetUpStartTimeDelay()/1000;
			}
			else
				failedupcount++;
			client->SetUploadState(US_NONE);
			client->ClearUploadBlockRequests();
//			CKnownFile* reqfile = theApp.sharedfiles->GetFileByID(client->GetUploadFileID());
//			if( reqfile )
//				reqfile->SubQueuedCount();
//			client->SetUploadFileID(NULL);
			return true;
		}
	}
	return false;
}

uint32 CUploadQueue::GetAverageUpTime(){
	if( successfullupcount ){
		return totaluploadtime/successfullupcount;
	}
	return 0;
}

bool CUploadQueue::RemoveFromWaitingQueue(CUpDownClient* client, bool updatewindow){
	POSITION pos = waitinglist.Find(client);
	if (pos){
		RemoveFromWaitingQueue(pos,updatewindow);
		if (updatewindow)
			theApp.emuledlg->transferwnd.ShowQueueCount(waitinglist.GetCount());
		return true;
	}
	else
		return false;
}

void CUploadQueue::RemoveFromWaitingQueue(POSITION pos, bool updatewindow){	
	CUpDownClient* todelete = waitinglist.GetAt(pos);
	waitinglist.RemoveAt(pos);
	if( todelete->IsBanned() )
		todelete->UnBan();
	if (updatewindow)
		theApp.emuledlg->transferwnd.queuelistctrl.RemoveClient(todelete);
	todelete->SetUploadState(US_NONE);
}


bool CUploadQueue::CheckForTimeOver(CUpDownClient* client){
	if( client->GetUpStartTimeDelay() > 3600000 ){ // Try to keep the clients from downloading for ever.
		theApp.emuledlg->AddDebugLogLine(false, "%s: Upload session ended due to excessive time.", client->GetUserName());
		return true;
	}
	// For some reason, some clients can continue to download after a chunk size.
	// Are they redownloading the same chunk over and over????
	if( client->GetSessionUp() > 10485760 ){
		theApp.emuledlg->AddDebugLogLine(false, "%s: Upload session ended due to excessive transfered amount.", client->GetUserName());
		return true;
	}

	if (!theApp.glob_prefs->TransferFullChunks()){

		// Cache current client score
		const uint32 score = client->GetScore(true, true);

		// Check if another client has a bigger score
		for(POSITION pos = waitinglist.GetHeadPosition(); pos != 0; )
			if(score < waitinglist.GetNext(pos)->GetScore(true, false)){
				theApp.emuledlg->AddDebugLogLine(false, "%s: Upload session ended due to score.", client->GetUserName());
				return true;
			}
	}
	return false;
}

void CUploadQueue::DeleteAll(){
	waitinglist.RemoveAll();
	uploadinglist.RemoveAll();
}

uint16 CUploadQueue::GetWaitingPosition(CUpDownClient* client){
	if (!IsOnUploadQueue(client))
		return 0;
	uint16 rank = 1;
	uint32 myscore = client->GetScore(false);
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;waitinglist.GetNext(pos)){
		if (waitinglist.GetAt(pos)->GetScore(false) > myscore)
			rank++;
	}
	return rank;
}

void CUploadQueue::CompUpDatarateOverhead(){
	// Patch by BadWolf - Accurate datarate Calculation
	TransferredData newitem = {m_nUpDataRateMSOverhead,::GetTickCount()};
	this->m_AvarageUDRO_list.AddTail(newitem);
	sumavgUDRO += m_nUpDataRateMSOverhead;
	while ((float)(m_AvarageUDRO_list.GetTail().timestamp - m_AvarageUDRO_list.GetHead().timestamp)  > MAXAVERAGETIME)
		sumavgUDRO -= m_AvarageUDRO_list.RemoveHead().datalen;
	m_nUpDataRateMSOverhead = 0;
	if(m_AvarageUDRO_list.GetCount() > 10)
		m_nUpDatarateOverhead = 1000 * sumavgUDRO /(m_AvarageUDRO_list.GetTail().timestamp - m_AvarageUDRO_list.GetHead().timestamp);
	else
		m_nUpDatarateOverhead = 0;
	return;
	// END Patch by BadWolf	
}

VOID CALLBACK CUploadQueue::UploadTimer(HWND hwnd, UINT uMsg,UINT_PTR idEvent,DWORD dwTime)
{
	// NOTE: Always handle all type of MFC exceptions in TimerProcs - otherwise we'll get mem leaks
	try
	{
		// Barry - Don't do anything if the app is shutting down - can cause unhandled exceptions
		if (!theApp.emuledlg->IsRunning())
			return;

		theApp.uploadqueue->Process();
		theApp.downloadqueue->Process();
		theApp.uploadqueue->CompUpDatarateOverhead();
		theApp.downloadqueue->CompDownDatarateOverhead();
		counter++;

		// one second
		if (counter >= 10){
			counter=0;
			theApp.clientcredits->Process();
			theApp.serverlist->Process();
			theApp.friendlist->Process();
			if( theApp.serverconnect->IsConnecting() && !theApp.serverconnect->IsSingleConnect() )
				theApp.serverconnect->TryAnotherConnectionrequest();

			theApp.emuledlg->statisticswnd.UpdateConnectionsStatus();
			if (theApp.glob_prefs->WatchClipboard4ED2KLinks()) theApp.emuledlg->searchwnd.SearchClipBoard();		
			if (theApp.serverconnect->IsConnecting()) theApp.serverconnect->CheckForTimeout();

			// -khaos--+++> Update connection stats...
			iupdateconnstats++;
			// 2 seconds
			if (iupdateconnstats>=2) {
				iupdateconnstats=0;
				theApp.emuledlg->statisticswnd.UpdateConnectionStats((float)theApp.uploadqueue->GetDatarate()/1024, (float)theApp.downloadqueue->GetDatarate()/1024);
			}
			// <-----khaos-

			// display graphs
			if (theApp.glob_prefs->GetTrafficOMeterInterval()>0) {
				igraph++;

				if (igraph >= (uint32)(theApp.glob_prefs->GetTrafficOMeterInterval()) ) {
					igraph=0;
					//theApp.emuledlg->statisticswnd.SetCurrentRate((float)(theApp.uploadqueue->Getavgupload()/theApp.uploadqueue->Getavg())/1024,(float)(theApp.uploadqueue->Getavgdownload()/theApp.uploadqueue->Getavg())/1024);
					theApp.emuledlg->statisticswnd.SetCurrentRate((float)(theApp.uploadqueue->GetDatarate())/1024,(float)(theApp.downloadqueue->GetDatarate())/1024);
					//theApp.uploadqueue->Zeroavg();
				}
			}
			if (theApp.emuledlg->activewnd == &theApp.emuledlg->statisticswnd && theApp.emuledlg->IsWindowVisible() )  {
				// display stats
				if (theApp.glob_prefs->GetStatsInterval()>0) {
					istats++;

					if (istats >= (uint32)(theApp.glob_prefs->GetStatsInterval()) ) {
						istats=0;
						theApp.emuledlg->ShowStatistics();
					}
				}
			}
			//save rates every second
			theApp.emuledlg->statisticswnd.RecordRate();

			sec++;
			// 5 seconds
			if (sec>=5) {

				#ifdef _DEBUG
					if (!AfxCheckMemory()) AfxDebugBreak();
				#endif

				sec = 0;
				theApp.listensocket->Process();
				theApp.OnlineSig(); // Added By Bouc7 
				theApp.emuledlg->ShowTransferRate();
				
				// update cat-titles with downloadinfos only when needed
				if (theApp.glob_prefs->ShowCatTabInfos() && 
					theApp.emuledlg->activewnd==&theApp.emuledlg->transferwnd && 
					theApp.emuledlg->IsWindowVisible()) 
						theApp.emuledlg->transferwnd.UpdateCatTabTitles();
				
				if (theApp.glob_prefs->IsSchedulerEnabled()) theApp.scheduler->Check();
			}

			statsave++;
			// -khaos--+++>
			// 60 seconds
			if (statsave>=60) {
				// Time to save our cumulative statistics.
				statsave=0;
				
				// CPreferences::SaveStats() in Preferences.cpp
				// This function does NOT update the tree!
				theApp.glob_prefs->SaveStats();
			// <-----khaos-
			}
		}
	}
	CATCH_DFLT_EXCEPTIONS("CUploadQueue::UploadTimer")
}

CUpDownClient* CUploadQueue::GetNextClient(CUpDownClient* lastclient){
	if (waitinglist.IsEmpty())
		return 0;
	if (!lastclient)
		return waitinglist.GetHead();
	POSITION pos = waitinglist.Find(lastclient);
	if (!pos){
		TRACE("Error: CServerList::GetNextClient");
		return waitinglist.GetHead();
	}
	waitinglist.GetNext(pos);
	if (!pos)
		return NULL;
	else
		return waitinglist.GetAt(pos);
}

void CUploadQueue::FindSourcesForFileById(CTypedPtrList<CPtrList, CUpDownClient*>* srclist, uchar* filehash) {
	POSITION pos;
	
	pos = uploadinglist.GetHeadPosition();
	while(pos) {
		CUpDownClient *potential = uploadinglist.GetNext(pos);
		if(md4cmp(potential->GetUploadFileID(), filehash) == 0)
			srclist->AddTail(potential);
	}

	pos = waitinglist.GetHeadPosition();
	while(pos) {
		CUpDownClient *potential = waitinglist.GetNext(pos);
		if(md4cmp(potential->GetUploadFileID(), filehash) == 0)
			srclist->AddTail(potential);
	}
}
